import 'dart:convert';
import 'dart:io';
import 'dart:developer' as developer;

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:insurance_card_scanner/models/scanned_card.dart';
import 'package:insurance_card_scanner/models/card_mapper.dart';
import 'package:insurance_card_scanner/models/card_response_model.dart';
import 'package:flutter_inappwebview/flutter_inappwebview.dart';

const captureConsole = '''
    var oldLog = console.log;
    console.log = function (message) {
      const _message = {
        command: 'console.log',
        message: message
      }
      window.flutter_inappwebview.callHandler('messageHandler', JSON.stringify(_message));
      oldLog.apply(console, arguments);
    };

    var oldDebug = console.debug;
    console.debug = function (message) {
      const _message = {
        command: 'console.debug',
        message: message
      }
      window.flutter_inappwebview.callHandler('messageHandler', JSON.stringify(_message));
      oldDebug.apply(console, arguments);
    };

    var oldError = console.error;
    console.error = function (message) {
      const _message = {
        command: 'console.error',
        message: message
      }
      window.flutter_inappwebview.callHandler('messageHandler', JSON.stringify(_message));
      oldError.apply(console, arguments);
    };
''';

typedef ScanErrorCallback = Function(String?);
typedef ScanSuccessCallback = Function(ScannedCard);

class CardScanner extends StatefulWidget {
  final String sessionToken;
  final ScanSuccessCallback onCardScanSuccess;
  final VoidCallback? onCardScanCancel;
  final ScanErrorCallback? onCardScanError;
  final VoidCallback? onWebViewLoadError;
  final VoidCallback? onWebViewCreated;
  final bool live;
  final bool debug;

  const CardScanner({
    Key? key,
    required this.sessionToken,
    required this.onCardScanSuccess,
    this.onCardScanCancel,
    this.onCardScanError,
    this.onWebViewLoadError,
    this.onWebViewCreated,
    this.live = false,
    this.debug = false,
  }) : super(key: key);

  @override
  State<CardScanner> createState() => _CardScannerState();
}

class _CardScannerState extends State<CardScanner> {
  final GlobalKey webViewKey = GlobalKey();
  bool hasInitalized = false;

  String get webViewUrl => widget.live
      ? 'https://api.cardscan.ai/v1/webview'
      : 'https://sandbox.cardscan.ai/v1/webview';

  InAppWebViewController? webViewController;
  InAppWebViewGroupOptions options = InAppWebViewGroupOptions(
      crossPlatform: InAppWebViewOptions(
        useShouldOverrideUrlLoading: true,
        mediaPlaybackRequiresUserGesture: false,
      ),
      android: AndroidInAppWebViewOptions(
        useHybridComposition: true,
      ),
      ios: IOSInAppWebViewOptions(
          allowsInlineMediaPlayback: true, allowsLinkPreview: false));

  void processMessage(payload) {
    try {
      final message = jsonDecode(payload);
      switch (message['command']) {
        case 'cancel':
          _log('cardScanCancel');
          widget.onCardScanCancel?.call();
          break;
        case 'success':
          _log('cardScanSuccess');
          final cardJson = message['card'];
          final card = cardJson is Map
              ? CardResponseModel.fromJson(cardJson).toCard()
              : null;
          if (card != null) {
            widget.onCardScanSuccess.call(card);
          } else {
            widget.onCardScanError?.call(null);
          }
          break;
        case 'error':
          final error = message['error'];
          final errorMessage = error is Map ? error['message'] : null;
          _log("cardScanError - $error");
          widget.onCardScanError?.call(errorMessage);
          break;
        case 'syserror':
        case 'console.log':
        case 'console.error':
        case 'console.debug':
          var error = message['message'];
          _log("WebView - $error");
          break;
        default:
          _log("Unknown command received from JS - $message");
      }
    } catch (e) {
      _log("Failed to parse JavascriptMessage - ${payload.message} $e");
    }
  }

  Future<void> enableWebviewDebug() async {
    if (Platform.isAndroid) {
      await AndroidInAppWebViewController.setWebContentsDebuggingEnabled(true);
    }
  }

  @override
  Widget build(BuildContext context) {
    final osName = Platform.operatingSystem.toLowerCase();
    var version = Uri.encodeComponent(Platform.operatingSystemVersion);
    var package = '0.1.2';
    final config = {
      "live": widget.live,
      "debug": widget.debug,
      "sessionToken": widget.sessionToken,
      //content: content,
      "enableCameraPermissionModal": false,
    };

    if (widget.debug) {
      enableWebviewDebug();
    }

    return InAppWebView(
        key: webViewKey,
        initialUrlRequest: URLRequest(
            url: Uri.parse(
                "$webViewUrl?platform=flutter-inapp&os=$osName&version=$version&package=$package")),
        initialOptions: options,
        onWebViewCreated: (controller) {
          webViewController = controller;
          _log("onWebViewCreated");

          controller.addJavaScriptHandler(
              handlerName: 'messageHandler',
              callback: (args) {
                _log("messageHandler: $args");
                for (var message in args) {
                  processMessage(message);
                }
              });

          controller.addJavaScriptHandler(
              handlerName: 'CardScanViewLoaded',
              callback: (args) async {
                _log("CardScanViewLoaded: $args");

                final javaScript = '''
                  if (window.loadCardScanView != undefined) {
                    window.loadCardScanView('${jsonEncode(config)}');
                  } else {
                    window.addEventListener('CardScanViewLoaded', function (e) {
                      window.loadCardScanView('${jsonEncode(config)}');
                    });
                  }

                  $captureConsole

                  //setup error monitoring
                  window.addEventListener('error', function (e) {
                    const message = {
                      command: 'syserror',
                      message: e.message
                    }
                    window.flutter_inappwebview.callHandler('messageHandler', JSON.stringify(message));
                  });

                  console.debug("injectedJavaScript -- complete");
                ''';

                try {
                  controller.evaluateJavascript(source: javaScript);
                } catch (e) {
                  //fails on some weird edge cases if native webview is in parital load/unload state.
                  developer.log(
                    'Error: Failed to init CardScanView',
                    name: widget.toString(),
                    error: jsonEncode(e),
                  );
                }
              });
        },
        androidOnPermissionRequest: (controller, origin, resources) async {
          return PermissionRequestResponse(
              resources: resources,
              action: PermissionRequestResponseAction.GRANT);
        },
        onLoadStop: (controller, url) {
          _log("onLoadStop: $url");
          //We cannot load the JS here as onLoadStop is called multiple times when our URL redirects.
          //When the webview is in the process of unloading this will get stuck and cause an error.
        },
        onLoadError: (controller, url, code, message) {
          _log("onLoadError: $code - $message");
          widget.onWebViewLoadError?.call();
        });
  }

  @override
  void dispose() {
    super.dispose();
    _log("dispose");
    //This seemed to clean up a state change bug.
    webViewController?.removeJavaScriptHandler(handlerName: 'messageHandler');
    webViewController?.removeJavaScriptHandler(
        handlerName: 'CardScanViewLoaded');
    webViewController = null;
  }

  void _log(String message) {
    if (kDebugMode) {
      debugPrint('CardScanModal $message');
    }
  }
}
